
# -*- encoding: utf-8 -*-
##############################################################################
#
#    Acrisel LTD
#    Copyright (C) 2008- Acrisel (acrisel.com) . All Rights Reserved
#
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################
"""
 
About
=========
:synopsis:     commandline  handler for existing eventor db
:moduleauthor: Arnon Sela @ acrisel.com
:date:         Nov 26, 2016
:description:  cmd provides access to set of commands to manipulate eventor db.  
               This would be done after a run and before rerun.


Command Line Arguments:
----------------------

.. cmdoption::  <Fire Perimeter> 

      Fire perimeter from FIMT Schema (Feature Class)
    
.. cmdoption::  <Fire to Export> 

      List of fires to export (The fire number in the feature class above that you want to load)
      
Example to run on the command line::

    running_module.py //some/url some.other.argument 
    

Outputs:
-------------------
If you script creates any outputs delete this line here and replace it 
with a description of the outputs produced by your script.  If your script
does not produce any outputs delete the entire output section!

Dependencies:
-------------------
Does your script have any dependencies?  If it does then include them here.  Only
include dependencies like 3rd party modules that your script requires that are not 
part of its hosting project.


**History:**
-------------------

:Author: <who made the changes>
:Modification:
   - Describe here the changes that were made in bullet form.  Use
     a new dash character '-' to create a new entry!
:Date: <enter the date the changes were made>


API DOC:
===============     
"""
# TODO:
# handler main commands
